# Plugin to add Flatpak building functionality
#
# The code here is partially based on the code in kojid,
# which has the following license notice:
#
#    Koji is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation;
#    version 2.1 of the License.
#
#    This software is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this software; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#       Mike McLean <mikem@redhat.com>
#       Mike Bonnet <mikeb@redhat.com>
#
# Plugin author:
#       Owen Taylor <otaylor@redhat.com

import copy
from dataclasses import dataclass
from datetime import datetime, timezone
from functools import cached_property
import glob
import hashlib
import json
import os
import re
import shutil
from textwrap import dedent
from typing import Any, Dict, List, Optional

from flatpak_module_tools import Arch, ContainerSpec, PackageLocator
from koji import Enum
from koji.daemon import log_output, SCM
import koji.tasks
from koji.tasks import ServerExit
from koji.util import format_shell_cmd, isSuccess, parseStatus

from __main__ import BuildRoot


CONFIG_FILE = '/etc/kojid/plugins/flatpak.conf'


class Option(Enum):
    REQUIRED = 1


@dataclass
class FlatpakConfig:
    registry: Optional[str]
    authfile: Optional[str]

    @classmethod
    def read(cls):
        cp = koji.read_config_files([CONFIG_FILE])
        registry = cp.get("push", "registry", fallback=None)
        if registry == "":
            registry = None
        # Very basic check
        if registry and not registry.startswith("docker://"):
            raise koji.GenericError(
                f"{CONFIG_FILE}:push: registry '{registry}' should start with docker://"
            )

        if registry:
            registry = registry[len("docker://"):]

        if registry and registry.endswith("/"):
            registry = registry[:-1]

        authfile = cp.get("push", "authfile", fallback=None)
        if authfile == "":
            authfile = None

        return cls(registry=registry, authfile=authfile)

    def make_skopeo_copy(self, src: str, unprefixed_dest: str,
                         extra_args: Optional[List[str]] = None):
        cmd = ["skopeo", "copy"]
        if self.authfile:
            cmd.append(f"--authfile={self.authfile}")
        if extra_args:
            cmd += extra_args

        cmd.append(src)
        cmd.append(f"docker://{self.registry}/{unprefixed_dest}")

        return cmd


@dataclass
class FlatpakBuildArchResultPayload:
    brootid: str
    output: str
    manifest: str
    config: str
    rpmlist: str
    logs: List[str]
    unique_tag: str


@dataclass
class FlatpakBuildArchResult(FlatpakBuildArchResultPayload):
    task_id: int
    arch: str


@dataclass
class ResultFile:
    local_path: str

    @cached_property
    def contents(self):
        with open(self.local_path, "rb") as f:
            return f.read()

    @cached_property
    def digest(self):
        return hashlib.sha256(self.contents).hexdigest()

    @cached_property
    def size(self):
        return len(self.contents)

    # Having a single .json property of type Any would means that .json() doesn't get
    # caught by type checkers

    @cached_property
    def json_dict(self) -> Dict[str, Any]:
        result = json.loads(self.contents)
        assert isinstance(result, dict)
        return result

    @cached_property
    def json_list(self) -> List[Any]:
        result = json.loads(self.contents)
        assert isinstance(result, list)
        return result


# This function is duplicated in Bodhi, but we don't count on Bodhi using
# the exact same tags - it independently comes up with tags based on the
# version and release when pushing to the final registry.

def make_valid_container_tag(original_tag):
    """
    Create a valid container tag from an arbitrary string.

    Turns an arbitrary string (typically a koji 'version' or 'version-release' string),
    into a valid OCI registry tag. This operation isn't reversable - multiple
    tags might end up as the same thing. For example, 1.0_a5', '1.0^a5', and '1.0~a5'
    all end up as '1.0_a5'.

    Args:
        original_tag: str: an arbitrary string
    Returns:
        str: a valid OCI registry tag based on original tag
    """
    # https://github.com/opencontainers/distribution-spec/blob/main/spec.md#pulling-manifests
    #
    #    Throughout this document, <reference> as a tag MUST be at most 128 characters in length
    #    and MUST match the following regular expression
    #
    #  [a-zA-Z0-9_][a-zA-Z0-9._-]{0,127}

    # Replace invalid characters with _
    new_tag = re.sub(r'^[^a-zA-Z0-9_]|[^a-zA-Z0-9._-]', '_', original_tag)

    # Truncate to 128 characters
    return new_tag[0:128]


class FlatpakBuildTask(koji.tasks.BaseTaskHandler):
    Methods = ['flatpakBuild']
    # we mostly just wait on other tasks
    _taskWeight = 0.2

    def handler(self, src: str, target, opts=None):
        """Handler for building Flatpaks"""
        if opts is None:
            opts = {}
        self.opts = opts
        if opts.get('arch_override') and not opts.get('scratch'):
            raise koji.BuildError("arch_override is only allowed for scratch builds")

        self.config = FlatpakConfig.read()

        task_info = self.session.getTaskInfo(self.id)
        target_info = None
        if target:
            target_info = self.session.getBuildTarget(target)
        if target_info:
            dest_tag_id = target_info['dest_tag']
            self.build_tag_id = target_info['build_tag']
            self.build_tag_name = target_info['build_tag_name']
        else:
            raise koji.GenericError('unknown build target: %s' % target)

        self.build_config = self.session.getBuildConfig(self.build_tag_id)

        # Can only build from SCM
        if not SCM.is_scm_url(src):
            raise koji.BuildError("src is not a SCM url")
        # policy checks...
        policy_data = {
            'user_id': task_info['owner'],
            'source': src,
            'task_id': self.id,
            'build_tag': self.build_tag_id,
            'skip_tag': bool(self.opts.get('skip_tag')),
            'scratch': opts.get('scratch'),
            'target': target_info['name'],
            'from_scm': SCM.is_scm_url(src),
        }
        if not self.opts.get('skip_tag'):
            policy_data['tag'] = dest_tag_id
        self.session.host.assertPolicy('build_rpm', policy_data)

        self.repo_info = self.getRepo(self.build_tag_id)  # (subtask)
        self.event_id = self.session.getLastEvent()['id']

        self.container_spec, source = self.checkout_container_spec(src)

        if self.container_spec.flatpak.build_runtime or self.container_spec.flatpak.build_extension:
            child_opts: Dict[str, Any] = self.get_runtime_opts()
        else:
            child_opts = self.get_app_opts()

        archlist = self.get_arch_list()
        data = self.get_name_version_data(child_opts, archlist)

        build_extra: Dict[str, Any] = {
            'image': {
                'flatpak': True,
            },
            'source': {
                'original_url': src
            }
        }

        data['task_id'] = self.id
        data['source'] = source['source']
        data['extra'] = build_extra

        self.logger.info("Reading package config for %(name)s" % data)
        pkg_cfg = self.session.getPackageConfig(dest_tag_id, data['name'], event=self.event_id)
        self.logger.debug("%r" % pkg_cfg)
        if not self.opts.get('skip_tag') and not self.opts.get('scratch'):
            # Make sure package is on the list for this tag
            if pkg_cfg is None:
                raise koji.BuildError("package %s not in list for tag %s"
                                      % (data['name'], target_info['dest_tag_name']))
            elif pkg_cfg['blocked']:
                raise koji.BuildError("package %s is blocked for tag %s"
                                      % (data['name'], target_info['dest_tag_name']))

        # let the system know about the build we're attempting
        if not self.opts.get('scratch'):
            self.logger.debug("Calling initImageBuild(%r, %r)", self.id, data)
            build_info = self.session.host.initImageBuild(self.id, data)
            nvr = "{name}-{version}-{release}".format(**build_info)
        else:
            # scratch builds do not get imported
            build_info = None
            nvr = "{name}-{version}-1".format(**data)

        try:
            child_opts['nvr'] = nvr
            child_opts['repo_id'] = self.repo_info['id']

            subtask_results = self.run_builds(archlist, child_opts=child_opts)

            index_metadata = self.push(nvr, subtask_results)

            results = {
                arch: self.image_data_for_subtask_result(subtask_result)
                for arch, subtask_result in subtask_results.items()
            }

            if build_info:
                self.session.host.completeImageBuild(self.id, build_info['id'], results)
            else:
                # scratch builds do not get imported
                self.session.host.moveImageBuildToScratch(self.id, results)
        except (SystemExit, ServerExit, KeyboardInterrupt):
            # we do not trap these
            raise
        except Exception:
            if build_info:
                # scratch builds do not get imported
                self.session.host.failBuild(self.id, build_info['id'])
            # reraise the exception
            raise

        if build_info:
            if index_metadata:
                # For testing purposes, we don't want to mutate the build_extra
                # dict, since it is saved in call args
                build_extra = copy.deepcopy(build_extra)
                build_extra["image"]["index"] = index_metadata

            archive_extra = {
                subtask_result.output: self.archive_extra_for_subtask_result(nvr, subtask_result)
                for subtask_result in subtask_results.values()
            }

            self.session.flatpakSetExtra(self.id, build_info['id'], build_extra, archive_extra)

        if build_info and not self.opts.get('skip_tag'):
            self.tag_build(build_info['id'], dest_tag_id)

    def checkout_container_spec(self, url: str):
        # will throw a BuildError if the url is invalid
        scm = SCM(url, allow_password=self.options.allow_password_in_scm_url)
        scm.assert_allowed(allowed=self.options.allowed_scms,
                           session=self.session,
                           by_config=self.options.allowed_scms_use_config,
                           by_policy=self.options.allowed_scms_use_policy,
                           policy_data={
                               'user_id': self.taskinfo['owner'],
                               'channel': self.session.getChannel(self.taskinfo['channel_id'],
                                                                  strict=True)['name'],
                               'scratch': self.opts.get('scratch')
                           })

        scmdir = self.workdir
        logfile = self.workdir + '/checkout.log'
        uploadpath = self.getUploadDir()

        self.run_callbacks('preSCMCheckout', scminfo=scm.get_info(),
                           build_tag=self.build_tag_id,
                           scratch=self.opts.get('scratch'))

        # Check out spec file, etc. from SCM
        sourcedir = scm.checkout(scmdir, self.session, uploadpath, logfile)
        self.run_callbacks("postSCMCheckout",
                           scminfo=scm.get_info(),
                           build_tag=self.build_tag_id,
                           scratch=self.opts.get('scratch'),
                           srcdir=sourcedir)

        container_yaml_path = os.path.join(sourcedir, "container.yaml")

        container_spec = ContainerSpec(container_yaml_path)
        self.uploadFile(container_yaml_path)

        return container_spec, scm.get_source()

    def get_extra_repo(self, extra_build_tag: str, *, dist=False):
        if extra_build_tag == self.build_config['name'] and not dist:
            return self.repo_info['id']
        else:
            if dist:
                # We can't easily wait for a repo to be created, which self.getRepo()
                # does for the non-dist-repo case, but that's a edge case anyways compared
                # to the case where the repo exists but is stale.
                extra_repo_info = self.session.getRepo(extra_build_tag, dist=True)
                if extra_repo_info is None:
                    raise koji.BuildError(f"No dist-repo exists for {extra_build_tag}")
            else:
                extra_repo_info = self.getRepo(extra_build_tag)  # (subtask)

            return extra_repo_info['id']

    def get_build_config_extra(self, key: str, default: Any = Option.REQUIRED):
        result = self.build_config['extra'].get(key)
        if result is None:
            if default == Option.REQUIRED:
                raise koji.BuildError(
                    f"Build tag '{self.build_config['name']}' doesn't have {key} set"
                )
            else:
                return default

        return result

    def get_runtime_opts(self):
        runtime_package_tag = self.get_build_config_extra('flatpak.runtime_package_tag')

        return {
            'runtime_rpm_repo_id': self.get_extra_repo(runtime_package_tag),
        }

    def get_app_opts(self):
        flatpak_spec = self.container_spec.flatpak

        runtime_tag = self.get_build_config_extra('flatpak.runtime_tag')
        runtime_package_tag = self.get_build_config_extra('flatpak.runtime_package_tag')
        runtime_package_dist_repo = self.get_build_config_extra(
            'flatpak.runtime_package_dist_repo', False
        )
        app_package_tag = self.get_build_config_extra('flatpak.app_package_tag')
        app_package_dist_repo = self.get_build_config_extra(
            'flatpak.app_package_dist_repo', False
        )

        tagged_builds = self.session.listTagged(
            runtime_tag, package=flatpak_spec.runtime_name, latest=True, inherit=True,
        )
        if len(tagged_builds) == 0:
            raise koji.BuildError(
                f"Can't find build for {flatpak_spec.runtime_name} in {runtime_tag}"
            )

        runtime_nvr = tagged_builds[0]["nvr"]
        runtime_version = tagged_builds[0]["version"]
        if flatpak_spec.runtime_version and flatpak_spec.runtime_version != runtime_version:
            raise koji.BuildError(
                f"Runtime '{runtime_nvr}' doesn't match "
                f"'runtime_version: {flatpak_spec.runtime_version}' in container.yaml"
            )

        return {
            'runtime_nvr':  runtime_nvr,
            'runtime_rpm_repo_id': self.get_extra_repo(
                runtime_package_tag, dist=runtime_package_dist_repo
            ),
            'app_rpm_repo_id': self.get_extra_repo(
                app_package_tag, dist=app_package_dist_repo
            ),
        }

    def get_arch_list(self):
        # get list of arches to build for
        arches = self.build_config['arches']
        if not arches:
            # XXX - need to handle this better
            raise koji.BuildError("No arches for tag %(name)s [%(id)s]" % self.build_config)
        assert isinstance(arches, str)
        tag_arches = {koji.canonArch(a) for a in arches.split()}
        self.logger.debug('tag arches: %s' % arches)
        result = set(arches.split())
        self.logger.debug('base archlist: %r' % result)

        result = {
            a for a in result if self.container_spec.platforms.includes_platform(a)
        }
        self.logger.debug('archlist after applying platforms from container.yaml: %r' % result)

        override = self.opts.get('arch_override')
        if self.opts.get('scratch') and override:
            # only honor override for scratch builds
            self.logger.debug('arch override: %s' % override)
            result = set(override.split())

        result = {
            a for a in result if koji.canonArch(a) in tag_arches
        }

        if not result:
            raise koji.BuildError("No matching arches were found")

        self.logger.debug("Final archlist: %r", result)

        return sorted(result)

    def get_name_version_data(self, child_opts: Dict[str, Any], archlist: List[str]):
        if self.container_spec.flatpak.build_runtime or self.container_spec.flatpak.build_extension:
            name = self.container_spec.flatpak.component or self.container_spec.flatpak.name
            version = self.container_spec.flatpak.branch

            if not name or not version:
                raise koji.BuildError(
                    "flatpak:name and flatpak:branch must be specified for a runtime or extension")

        else:
            main_package = self.container_spec.flatpak.packages[0].name
            # Calling out to the server to get this information that we already had earlier
            # is a bit lame, but avoids code reorganization.
            repo = self.session.repoInfo(child_opts["app_rpm_repo_id"])
            pathinfo = koji.PathInfo(topdir=self.options.topurl)

            if repo["dist"]:
                baseurl = pathinfo.distrepo(repo["id"], repo["tag_name"], None) + "/$basearch/"
            else:
                baseurl = pathinfo.repo(repo["id"], repo["tag_name"]) + "/$basearch/"

            locator = PackageLocator()
            locator.add_repo(baseurl)
            latest_version = locator.find_latest_version(main_package, arch=Arch(rpm=archlist[0]))
            if not latest_version:
                raise koji.BuildError(
                    f"Can't find build for {main_package} in {repo['tag_name']}"
                )

            name = self.container_spec.flatpak.get_component_label(main_package)
            version = latest_version.version

        return {
            "name": name,
            "epoch": 0,
            "version": version,
        }

    def run_builds(self, archlist: List[str], child_opts):
        container_yaml = self.getUploadDir() + "/" + "container.yaml"
        self.logger.debug("Spawning jobs for arches: %r" % (archlist))
        subtasks: Dict[str, int] = {}
        for arch in archlist:
            arglist = [container_yaml, self.build_tag_id, arch, child_opts]
            subtasks[arch] = self.session.host.subtask(method='flatpakBuildArch',
                                                       arglist=arglist,
                                                       label=arch,
                                                       parent=self.id,
                                                       arch=arch,
                                                       opts=child_opts)

        self.logger.debug("Got subtasks: %r" % (subtasks))
        self.logger.debug("Waiting on subtasks...")

        # wait for subtasks to finish. With all=True, failany=True, it will
        # cancel other tasks and raise an exception if any task fails.
        subtask_results = self.wait(list(subtasks.values()), all=True, failany=True)
        return {
            arch: FlatpakBuildArchResult(task_id=task_id, arch=arch, **subtask_results[task_id])
            for (arch, task_id) in subtasks.items()
        }

    def get_result_file(self, subtask_result: FlatpakBuildArchResult, subtask_path):
        local_path = self.localPath(os.path.join(
            "work", koji.pathinfo.taskrelpath(subtask_result.task_id), subtask_path
        ))
        return ResultFile(local_path)

    def image_data_for_subtask_result(self, subtask_result: FlatpakBuildArchResult):
        rpmlist = self.get_result_file(subtask_result, subtask_result.rpmlist).json_list

        return {
            'task_id': subtask_result.task_id,
            'arch': subtask_result.arch,
            'files': [subtask_result.output],
            'logs': subtask_result.logs,
            'rpmlist': rpmlist,
        }

    def archive_extra_for_subtask_result(self, nvr, subtask_result: FlatpakBuildArchResult):
        config = self.get_result_file(subtask_result, subtask_result.config)
        manifest = self.get_result_file(subtask_result, subtask_result.manifest)

        layer_sizes = [{
            "diff_id": diff_id,
            "size": layer["size"]
        } for layer, diff_id in zip(
            manifest.json_dict["layers"], config.json_dict["rootfs"]["diff_ids"]
        )]

        name, version, release = nvr.rsplit("-", 2)
        name_label = self.container_spec.flatpak.get_name_label(name)

        result = {
            'image': {
                'arch': subtask_result.arch
            },
            'docker': {
                'config': config.json_dict,
                'id': manifest.json_dict["config"]["digest"],
                'layer_sizes': layer_sizes,
                'digests': {
                    "application/vnd.oci.image.manifest.v1+json": f"sha256:{manifest.digest}"
                }
            }
        }

        if self.config.registry:
            result["docker"].update({
                'repositories': [
                    f"{self.config.registry}/{name_label}@sha256:{manifest.digest}",
                    f"{self.config.registry}/{name_label}:{subtask_result.unique_tag}",
                ],
                'tags': [
                    subtask_result.unique_tag
                ]
            })

        return result

    def push(self, nvr: str, subtask_results: Dict[str, FlatpakBuildArchResult]):
        if not self.config.registry:
            return None

        def manifest_descriptor(subtask_result: FlatpakBuildArchResult):
            manifest = self.get_result_file(subtask_result, subtask_result.manifest)
            config = self.get_result_file(subtask_result, subtask_result.config)
            return {
                "mediaType": "application/vnd.oci.image.manifest.v1+json",
                "digest": f"sha256:{manifest.digest}",
                "size": manifest.size,
                "platform": {
                    "architecture": config.json_dict["architecture"],
                    "os": "linux",
                }
            }

        index = {
            "schemaVersion": 2,
            "mediaType": "application/vnd.oci.image.index.v1+json",
            "manifests": [
                 manifest_descriptor(subtask_result)
                 for subtask_result in subtask_results.values()
            ]
        }

        index_contents = json.dumps(index, indent=4).encode("utf8")
        index_contents_digest = hashlib.sha256(index_contents).hexdigest()

        image_dir = os.path.join(self.workdir, "index_image")
        koji.ensuredir(image_dir)

        blobdir = os.path.join(image_dir, "blobs/sha256")
        koji.ensuredir(blobdir)
        with open(os.path.join(blobdir, index_contents_digest), "wb") as f:
            f.write(index_contents)

        archive_index = {
            "schemaVersion": 2,
            "mediaType": "application/vnd.oci.image.index.v1+json",
            "manifests": [{
                "mediaType": "application/vnd.oci.image.index.v1+json",
                "digest": f"sha256:{index_contents_digest}",
                "size": len(index_contents),
                "annotations": {
                    "org.opencontainers.image.ref.name": nvr
                }
            }]
        }

        with open(os.path.join(image_dir, "index.json"), "w") as f:
            json.dump(archive_index, f, indent=4)

        name, version, release = nvr.rsplit("-", 2)
        name_label = self.container_spec.flatpak.get_name_label(name)

        version_tag = make_valid_container_tag(version)
        version_release_tag = make_valid_container_tag(f"{version}-{release}")

        for tag in [version_release_tag, version_tag, "latest"]:
            cmd = self.config.make_skopeo_copy(f"oci:{image_dir}",
                                               f"{name_label}:{tag}",
                                               extra_args=["--multi-arch=index-only"])
            logfile = f"{self.workdir}/push.log"
            status = log_output(self.session, cmd[0], cmd, logfile, self.getUploadDir(),
                                logerror=True)
            if not isSuccess(status):
                raise koji.BuildError(
                    "Failed to upload image to registry: "
                    f"{parseStatus(status, format_shell_cmd(cmd))}"
                )

        return {
            "digests": {
                "application/vnd.oci.image.index.v1+json": f"sha256:{index_contents_digest}"
            },
            "floating_tags": [
                version_tag, "latest"
            ],
            "pull": [
                f"{self.config.registry}/{name_label}@sha256:{index_contents_digest}",
                f"{self.config.registry}/{name_label}:{version_release_tag}"
            ],
            "tags": [
                version_release_tag
            ],
            "unique_tags": [
            ],
        }

    def tag_build(self, build_id: int, dest_tag_id: int):
        task_id = self.session.host.subtask(method='tagBuild',
                                            arglist=[dest_tag_id, build_id, False, None, True],
                                            label='tag',
                                            parent=self.id,
                                            arch='noarch')
        self.wait(task_id)


class FlatpakBuildArchTask(koji.tasks.BaseTaskHandler):
    Methods = ['flatpakBuildArch']

    def handler(self, container_yaml: str, build_tag: int, arch: str,
                opts: Optional[Dict[str, Any]] = None):
        self.config = FlatpakConfig.read()

        local_container_yaml = self.localPath(f"work/{container_yaml}")
        if not os.path.exists(local_container_yaml):
            raise koji.BuildError("container.yaml file missing: %s" % local_container_yaml)

        self.logger.debug("local_container_yaml path: %r", local_container_yaml)

        container_spec = ContainerSpec(local_container_yaml)

        if opts is None:
            opts = {}

        nvr = opts.get('nvr', None)
        if not nvr:
            raise koji.BuildError("The NVR of the build must be provided")

        repo_id = opts.get('repo_id')
        if not repo_id:
            raise koji.BuildError("A repo id must be provided")

        rootopts = {
            'install_group': 'flatpak-build',
            'repo_id': repo_id,
            'setup_dns': True,
        }
        broot = BuildRoot(self.session, self.options, build_tag, arch, self.id, **rootopts)
        broot.workdir = self.workdir

        def run_mock(what, cmd):
            self.logger.debug("%s: %s", what, format_shell_cmd(cmd))
            rv = broot.mock(cmd)
            if rv:
                raise koji.BuildError(
                    f"{what} failed: {parseStatus(rv, format_shell_cmd(cmd))}"
                )

        self.logger.debug("Initializing buildroot")
        broot.init()

        koji_conf_path = os.path.join(self.workdir, "koji.conf")
        with open(koji_conf_path, "w") as f:
            f.write(dedent(f"""\
                [koji]
                server={self.options.server}
                topurl={self.options.topurl}
                """))

        run_mock("Copying koji.conf", [
            "--copyin", koji_conf_path, "/etc/koji.conf"
        ])

        koji.ensuredir(broot.tmpdir())
        shutil.copy(local_container_yaml, os.path.join(broot.tmpdir(), "container.yaml"))
        target_container_yaml = os.path.join(broot.tmpdir(within=True), "container.yaml")

        koji.ensuredir(os.path.join(broot.tmpdir(), "work"))
        target_workdir = os.path.join(broot.tmpdir(within=True), "work")

        options = [
            "--containerspec", target_container_yaml,
            "--workdir", target_workdir,
            "--resultdir", broot.tmpdir(within=True),
            "--nvr", nvr,
            "--runtime-repo", str(opts["runtime_rpm_repo_id"]),
        ]

        if not container_spec.flatpak.build_runtime and not container_spec.flatpak.build_extension:
            options += [
                "--runtime-nvr", opts["runtime_nvr"],
                "--app-repo", str(opts["app_rpm_repo_id"]),
            ]

        # We create the log file for the Flatpak build process as a hard-link
        # from the buildroot into the resultdir. Since it's in the resultdir
        # and matches *.log, it will be uploaded by the code in BuildRoot.mock().

        result_logfile = os.path.join(broot.resultdir(), "flatpak.log")
        with open(result_logfile, "w") as f:
            pass

        buildroot_logfile = os.path.join(broot.tmpdir(), "flatpak.log")
        os.link(result_logfile, buildroot_logfile)

        target_logfile = os.path.join(broot.tmpdir(within=True), "flatpak.log")

        run_mock("Creating Flatpak", [
            "--chroot", "--",
            "bash", "-o", "pipefail", "-c", f'$0 "$@" 2>&1 | tee -a {target_logfile}',
            "flatpak-module", "assemble"
        ] + options)

        base_filename = f"{nvr}.{arch}.oci"
        for f in os.listdir(broot.tmpdir()):
            if f.startswith(base_filename + ".tar"):
                tar_filename = f
                break
        else:
            raise koji.BuildError("Failed to find output tarfile")

        base_location = os.path.join(broot.tmpdir(), base_filename)
        tar_location = os.path.join(broot.tmpdir(), tar_filename)

        self.uploadFile(tar_location)
        self.uploadFile(base_location + ".manifest.json")
        self.uploadFile(base_location + ".config.json")

        # self.id is already unique - adding other fields here just makes
        # the tag more informative
        timestamp = datetime.now(timezone.utc).strftime('%Y%m%d%H%M%S')
        unique_tag = f"{broot.tag_name}-{self.id}-{arch}-{timestamp}"
        self.push(tar_location, base_location + ".config.json", unique_tag)

        if container_spec.flatpak.build_runtime or container_spec.flatpak.build_extension:
            content_repo_id = opts["runtime_rpm_repo_id"]
        else:
            content_repo_id = opts["app_rpm_repo_id"]

        processed_manifest_location = os.path.join(self.workdir, f"{nvr}.{arch}.oci.rpmlist.json")

        self.process_rpm_manifest(
            broot, content_repo_id,
            base_location + ".rpmlist.json",
            processed_manifest_location
        )
        self.uploadFile(processed_manifest_location)

        log_files = glob.glob('%s/*.log' % broot.resultdir())

        broot.expire()

        return FlatpakBuildArchResultPayload(
            brootid=broot.id,
            output=tar_filename,
            manifest=f"{nvr}.{arch}.oci.manifest.json",
            config=f"{nvr}.{arch}.oci.config.json",
            rpmlist=f"{nvr}.{arch}.oci.rpmlist.json",
            logs=[
                os.path.basename(fn) for fn in log_files
            ],
            unique_tag=unique_tag,
        ).__dict__

    def process_rpm_manifest(self, broot: BuildRoot, content_repo_id: int,
                             original_rpmlist_path: str, processed_rpmlist_path: str):
        with open(original_rpmlist_path, "r") as f:
            rpmlist = json.load(f)

        # This is a hack to reuse the complicated code in BuildRoot.markExternalRPMs -
        # we pretend that the repository used to install the content is
        # the repository used to install the buildroot. The actual repository for
        # the buildroot is only used for installing tooling in the buildroot.
        save_repo_info = broot.repo_info
        broot.repo_info = self.session.repoInfo(content_repo_id)
        broot.markExternalRPMs(rpmlist)
        broot.repo_info = save_repo_info

        with open(processed_rpmlist_path, "w") as f:
            json.dump(rpmlist, f, indent=4)

    def push(self, archive_path: str, config_path: str, unique_tag: str):
        if not self.config.registry:
            return None

        with open(config_path, "r") as f:
            config = json.load(f)
        name_label = config["config"]["Labels"]["name"]

        cmd = self.config.make_skopeo_copy(
            f"oci-archive:{archive_path}", f"{name_label}:{unique_tag}"
        )
        logfile = f"{self.workdir}/push.log"
        status = log_output(self.session, cmd[0], cmd, logfile, self.getUploadDir(), logerror=True)
        if not isSuccess(status):
            raise koji.BuildError(
                "Failed to upload image to registry: "
                f"{parseStatus(status, format_shell_cmd(cmd))}"
            )
